{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE OverloadedStrings #-}

module Shift.Rendering where

import CMarkGFM as M
import Control.Lens ((^.))
import Control.Monad.Catch (MonadThrow)
import Control.Monad.IO.Class (MonadIO, liftIO)
import Control.Monad.State (MonadState)
import Control.Monad.Trans.Writer (WriterT (runWriterT), execWriterT, runWriter)
import Control.Monad.Trans.Writer.Lazy (tell)
import Data.Git (RefName(refNameRaw), RefName(refNameRaw), Commit, Ref, commitAuthor)
import Data.List (sortOn)
import Data.Maybe (catMaybes)
import Data.Monoid ((<>))
import Data.String.Conversions (cs)
import Data.Text (Text)
import qualified Data.Text as T (pack, pack, take)
import qualified Data.Text.IO as TIO (putStr, putStrLn)
import Data.Versions (prettyV)
import Shift.Types

renderRef ::
  (MonadIO m, MonadState s m, ClientState s, MonadThrow m) =>
  Ref hash ->
  m [Node]
renderRef ref = do
  url <- getRefURL ref

  let shortRef = T.take 7 . cs . show $ ref
  let shortRefNode = leafNode (CODE shortRef)

  pure $ case url of
    Just url_ -> [textNode "[", parentNode (LINK url_ shortRef) [shortRefNode], textNode "]"]
    Nothing -> [textNode "[", shortRefNode, textNode "]"]

renderConventionalCommit ::
  (MonadIO m, MonadState s m, ClientState s, MonadThrow m) =>
  ConventionalGroup hash ->
  m Node
renderConventionalCommit (ref, commit, pc) = do
  renderedRef <- renderRef ref
  authorText <- renderAuthor commit ref

  pure $
    listItemNode
      [ paragraphNode $
          renderedRef
            <> catMaybes
              [ Just $ textNode " ",
                Just . bold $ pc ^. ccScope <> ":",
                Just $ textNode " ",
                Just . textNode $ pc ^. ccSubject,
                Just $ textNode " ",
                authorText
              ]
      ]

renderAuthor ::
  (MonadIO m, MonadState s m, ClientState s, MonadThrow m) =>
  Commit hash ->
  Ref hash ->
  m (Maybe Node)
renderAuthor commit ref = do
  authorInfo <- getAuthorInfo (commitAuthor commit) ref

  pure $ case authorInfo of
    Just (username, authorUrl) -> Just $ parentNode (LINK authorUrl username) [textNode username]
    Nothing -> Nothing

renderMiscCommit ::
  (MonadIO m, MonadState s m, ClientState s, MonadThrow m) =>
  MiscGroup hash ->
  m Node
renderMiscCommit (ref, commit, MiscCommit subject) = do
  renderedRef <- renderRef ref
  authorText <- renderAuthor commit ref

  pure $
    listItemNode
      [ paragraphNode $
          renderedRef
            <> catMaybes
              [ Just $ textNode subject,
                Just $ textNode " ",
                authorText
              ]
      ]

renderMergeCommit ::
  (MonadIO m, MonadState s m, ClientState s, MonadThrow m) =>
  MergeGroup hash ->
  m Node
renderMergeCommit (ref, commit, MergeCommit subject) = do
  renderedRef <- renderRef ref
  authorText <- renderAuthor commit ref

  pure $
    listItemNode
      [ paragraphNode $
          renderedRef
            <> catMaybes
              [ Just $ textNode $ "Merge " <> subject,
                authorText
              ]
      ]

renderRange :: TagRef -> TagRef -> Text
renderRange tx ty =
  mconcat
    [renderTagRef tx, " to ", renderTagRef ty]

renderTagRef :: TagRef -> Text 
renderTagRef tr = case _tVersioning tr of
  Just versioning -> prettyV versioning
  Nothing -> T.pack . refNameRaw $ _tRef tr

bold :: Text -> Node
bold x = parentNode STRONG [textNode x]

indented :: Int -> Text -> Text
indented levels x = foldl (\acc _ -> acc <> "  ") "" [1 .. levels] <> x

indentedL :: Int -> Text -> Text
indentedL levels x = "\\" <> foldl (\acc _ -> acc <> "-") "" [1 .. levels] <> x

headerOne :: Text -> Node
headerOne t = parentNode (HEADING 1) [textNode t]

headerTwo :: Text -> Node
headerTwo t = parentNode (HEADING 2) [textNode t]

headerThree :: Text -> Node
headerThree t = parentNode (HEADING 3) [textNode t]

textNode :: Text -> Node
textNode t = Node Nothing (TEXT t) []

leafNode :: NodeType -> Node
leafNode n = Node Nothing n []

parentNode :: NodeType -> [Node] -> Node
parentNode = Node Nothing

listItemNode :: [Node] -> Node
listItemNode = parentNode ITEM

paragraphNode :: [Node] -> Node
paragraphNode = parentNode PARAGRAPH

printReport ::
  (MonadIO m, MonadState s m, ClientState s, MonadThrow m) =>
  (TagRef, TagRef) ->
  ChangeReport hash ->
  m [Node]
printReport (rangeStart, rangeEnd) report = execWriterT $ do
  tell [headerOne $ renderRange rangeStart rangeEnd]

  -- Print conventional commits
  conventionalSection "New features:" crFeatures
  conventionalSection "Fixes:" crFixes
  conventionalSection "Updated documentation:" crDocs
  conventionalSection "Refactors:" crRefactors
  conventionalSection "Updated tests:" crTests
  conventionalSection "Style fixes:" crStyles
  conventionalSection "Chores:" crChores

  -- Print miscs and merges
  miscCommits <- mapM renderMiscCommit (report ^. crMisc)
  mergeCommits <- mapM renderMergeCommit (report ^. crMerges)

  textSection "Miscellaneous changes:" miscCommits
  textSection "Merge commits:" mergeCommits
  where
    conventionalSection label sectionLens = do
      let rawCommits = sortOn (\(_, _, pc) -> pc ^. ccScope) (report ^. sectionLens)

      commits <- mapM renderConventionalCommit rawCommits
      textSection label commits

    textSection label commits = case commits of
      [] -> pure ()
      _ -> do
        tell [headerTwo label]

        tell [parentNode (LIST (ListAttributes BULLET_LIST False 0 PAREN_DELIM)) commits]
